#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <float.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include "lr.h"
#ifndef min
template <class T> inline T min(T x,T y) { return (x<y)?x:y; }
#endif
#ifndef max
template <class T> inline T max(T x,T y) { return (x>y)?x:y; }
#endif
template <class T> inline void swap(T& x, T& y) { T t=x; x=y; y=t; }
template <class S, class T> inline void clone(T*& dst, S* src, int n)
{
	dst = new T[n];
	memcpy((void *)dst,(void *)src,sizeof(T)*n);
}
#define Malloc(type,n) (type *)malloc((n)*sizeof(type))
#if 1
void info(const char *fmt,...)
{
	va_list ap;
	va_start(ap,fmt);
	vprintf(fmt,ap);
	va_end(ap);
}
void info_flush()
{
	fflush(stdout);
}
#else
void info(char *fmt,...) {}
void info_flush() {}
#endif

#ifdef __cplusplus
extern "C" {
#endif
extern double dnrm2_(int *, double *, int *);
extern double ddot_(int *, double *, int *, double *, int *);
extern int daxpy_(int *, double *, double *, int *, double *, int *);
extern int dscal_(int *, double *, double *, int *);
#ifdef __cplusplus
}
#endif

void Xv(int l, double *x_val, long int *x_ir, long int *x_jc, double *v, double *Xv)
{
	int i, j;

	for(i=0;i<l;i++) 
	{
		Xv[i] = 0;
		for(j=x_jc[i];j<x_jc[i+1];j++)
			Xv[i] += v[x_ir[j]]*x_val[j];
	}

}
void XTv(int l, int n, double *x_val, long int *x_ir, long int *x_jc, double *v, double *XTv)
{
	int i, j;
	for(i=0;i<n;i++) XTv[i]=0;
	for(i=0;i<l;i++) 
		for(j=x_jc[i];j<x_jc[i+1];j++)
			XTv[x_ir[j]] += x_val[j]*v[i];
}


double fun(int l, int n, double *w, double *x_val, long int *x_ir, long int *x_jc, double *y, double *C, double *z)
// z cannot be changed after Xv
// but z is changed in grad
{
	int i;
	double f=0;

	XTv(n, l, x_val, x_ir, x_jc, w, z);
	for(i=0;i<l;i++) 
	{
	        double yz = y[i]*z[i];
		if (yz >= 0)
		        f += C[i]*log(1 + exp(-yz));
		else
		        f += C[i]*(-yz+log(1 + exp(yz)));
	}
	f = 2*f;
	for(i=0;i<n;i++) 
		f += w[i]*w[i];
	f /= 2.0;

	return(f);
	// later need better implementation to avoid numerical inaccuracy
}

void grad(int l, int n, double *w, double *x_val, long int *x_ir, long int *x_jc, double *y, double *C, double *z, double *g, double *D)
{
	int i;

	for(i=0;i<l;i++) 
	{
		z[i] = 1/(1 + exp(-y[i]*z[i]));
		D[i] = z[i]*(1-z[i]);
		z[i] = C[i]*(z[i]-1)*y[i];
	}
	Xv(n, x_val, x_ir, x_jc, z, g);	

	for(i=0;i<n;i++) 
		g[i] = w[i] + g[i];
}

void Hv(int l, int n, double *x_val, long int *x_ir, long int *x_jc, double *D, double *C, double *s, double *Hs)
{
	int i;
	double *wa = Malloc(double, l);

	for(i=0;i<l;i++) wa[i]=0;
	XTv(n, l, x_val, x_ir, x_jc, s, wa);
	for(i=0;i<l;i++) 
		wa[i] = C[i]*D[i]*wa[i];

	Xv(n, x_val, x_ir, x_jc, wa, Hs);	
	for(i=0;i<n;i++) 
		Hs[i] = s[i] + Hs[i];
	free(wa);
}

// l2svm

void subXv(int *I, int sizeI, int l, double *x_val, long int *x_ir, long int *x_jc, double *v, double *Xv)

{
	int i, j;

	for(i=0;i<sizeI;i++)
	{
		Xv[i]=0;
		for(j=x_jc[I[i]];j<x_jc[I[i]+1];j++)
			Xv[i] += v[x_ir[j]]*x_val[j];
	}
}

void subXTv(int *I, int sizeI, int l, int n, double *x_val, long int *x_ir, long int *x_jc, double *v, double *XTv)
{
	int i, j, t;
	for(i=0;i<n;i++) XTv[i]=0;
	for(t=0;t<sizeI;t++)
	{
		i = I[t];
		for(j=x_jc[i];j<x_jc[i+1];j++)
			XTv[x_ir[j]] += x_val[j]*v[t];
	}
}

double l2svmfun(int l, int n, double *w, double *x_val, long int *x_ir, long int *x_jc, double *y, double *C, double *z)
{
	int i;
	double f=0;

	XTv(n, l, x_val, x_ir, x_jc, w, z);
	for(i=0;i<l;i++)
	{
	        z[i] = y[i]*z[i];
		double d = z[i]-1;
		if (d < 0)
			f += C[i]*d*d;
	}
	f = 2*f;
	for(i=0;i<n;i++)
		f += w[i]*w[i];
	f /= 2.0;

	return(f);
}

int l2svmgrad(int *I, int l, int n, double *w, double *x_val, long int *x_ir, long int *x_jc, double *y, double *C, double *z, double *g, double *D)
{
	int i;

	int sizeI = 0;
	for (i=0;i<l;i++)
		if (z[i] < 1)
		{
			z[sizeI] = C[i]*y[i]*(z[i]-1);
			I[sizeI] = i;
			sizeI++;
		}
	subXTv(I, sizeI, l, n, x_val, x_ir, x_jc, z, g);	
	for(i=0;i<n;i++)
		g[i] = w[i] + 2*g[i];
	return(sizeI);
}

void l2svmHv(int *I, int sizeI, int l, int n, double *x_val, long int *x_ir, long int *x_jc, double *D, double *C, double *s, double *Hs)

{
	int i;
	double *wa = new double[l];

	subXv(I, sizeI, l, x_val, x_ir, x_jc, s, wa);
	for(i=0;i<sizeI;i++)
		wa[i] = C[I[i]]*wa[i];

	subXTv(I, sizeI, l, n, x_val, x_ir, x_jc, wa, Hs);	
	for(i=0;i<n;i++)
		Hs[i] = s[i] + 2*Hs[i];
	free(wa);
}


int trcg(int l, int n, double delta, double *g, double *D, double *C, double *x_val, long int *x_ir, long int *x_jc, double *y, double *s, double *r)
{
	int i, inc = 1;
	double one = 1;
	double *d = Malloc(double, n);
	double *Hd = Malloc(double, n);
	double rTr, rnewTrnew, alpha, beta, cgtol;

	for (i=0; i<n; i++)
	{
		s[i] = 0;
		r[i] = -g[i];
		d[i] = r[i];
	}
	cgtol = 0.1*dnrm2_(&n, g, &inc);

	int cg_iter = 0;
	rTr = ddot_(&n, r, &inc, r, &inc);
	while (1)
	{
		if (dnrm2_(&n, r, &inc) <= cgtol)
			break;
		cg_iter++;
		Hv(l, n, x_val, x_ir, x_jc, D, C, d, Hd);
		alpha = rTr/ddot_(&n, d, &inc, Hd, &inc);
		daxpy_(&n, &alpha, d, &inc, s, &inc);
		if (dnrm2_(&n, s, &inc) > delta)
		{
			info("cg reaches trust region boundary\n");
			alpha = -alpha;
			daxpy_(&n, &alpha, d, &inc, s, &inc);

			double std = ddot_(&n, s, &inc, d, &inc); 
			double sts = ddot_(&n, s, &inc, s, &inc);
			double dtd = ddot_(&n, d, &inc, d, &inc); 
			double dsq = delta*delta;
			double rad = sqrt(std*std + dtd*(dsq-sts));
			if (std >= 0)
				alpha = (dsq - sts)/(std + rad);
			else
				alpha = (rad - std)/dtd;
			daxpy_(&n, &alpha, d, &inc, s, &inc);
			alpha = -alpha; 
			daxpy_(&n, &alpha, Hd, &inc, r, &inc);		
			break;
		}
		alpha = -alpha; 
		daxpy_(&n, &alpha, Hd, &inc, r, &inc);		
		rnewTrnew = ddot_(&n, r, &inc, r, &inc);
		beta = rnewTrnew/rTr; 
		dscal_(&n, &beta, d, &inc);
		daxpy_(&n, &one, r, &inc, d, &inc);				
		rTr = rnewTrnew;
	}
	free(d);
	free(Hd);

	return(cg_iter);
}


double norm_inf(int n, double *x)
{
	double dmax = fabs(x[0]);
	for (int i=1; i<n; i++)
		if (fabs(x[i]) >= dmax) 
			dmax = fabs(x[i]);
	return(dmax);
}

static void tron(const lr_problem *prob, const lr_parameter *param, double *w, double Cp, double Cn) 
{
	/* Parameters for updating the iterates. */
	double eta0 = 1e-4, eta1 = 0.25, eta2 = 0.75;
	
	/* Parameters for updating the trust region size delta. */
	double sigma1 = 0.25, sigma2 = 0.5, sigma3 = 4;

	int l = prob->l, n = prob->n;
	int i, cg_iter;
	double eps = param->eps;
	info("eps %f Cp %f Cn %f\n", eps, Cp, Cn);
	double delta, snorm, one=1.0;
	double alpha, f, fnew, prered, actred, gs;
	int search = 1, iter = 1, inc = 1, max_iter;	
	double *x_val = prob->x_val;
	long int *x_ir = prob->x_ir;
	long int *x_jc = prob->x_jc;
	double *y = prob->y;
	double *s = Malloc(double, n);
	double *r = Malloc(double, n);
	double *w_new = Malloc(double, n);
	double *z = Malloc(double, l);
	double *g = Malloc(double, n);
	double *D = Malloc(double, l);
	double *C = Malloc(double, l);

	// for L2 SVM
	int *I = Malloc(int ,l);
	int sizeI;


	for (i=0; i<n; i++)
		w[i] = 0;
	for (i=0; i<l; i++)
	{
		if (y[i] == 1)
			C[i] = Cp;
		else
			C[i] = Cn;
	}

        f = fun(l, n, w, x_val, x_ir, x_jc, y, C, z);
	grad(l, n, w, x_val, x_ir, x_jc, y, C, z, g, D);
	delta = dnrm2_(&n, g, &inc);

	if (norm_inf(n, g) < eps)
		search = 0;

	max_iter = 1000;
	iter = 1;

	double start = clock();

	while (iter <= max_iter && search)
	{
		double g_norm = norm_inf(n, g);
		info("FUN %20.15e GRAD %20.15e TIME %20.15e\n", f, g_norm, (clock()-start)/CLOCKS_PER_SEC); 
		if (g_norm < eps)
			break;

		cg_iter = trcg(l, n, delta, g, D, C, x_val, x_ir, x_jc, y, s, r);

		memcpy(w_new, w, sizeof(double)*n);
		daxpy_(&n, &one, s, &inc, w_new, &inc);				

		gs = ddot_(&n, g, &inc, s, &inc);
		prered = -0.5*(gs - ddot_(&n, s, &inc, r, &inc));
                fnew = fun(l, n, w_new, x_val, x_ir, x_jc, y, C, z);		

		/* Compute the actual reduction. */
	        actred = f - fnew;                
                                    
		/* On the first iteration, adjust the initial step bound. */
		snorm = dnrm2_(&n, s, &inc);
		if (iter == 1)
			delta = min(delta, snorm);
			   
		/* Compute prediction alpha*snorm of the step. */
		if (fnew - f - gs <= 0)
			alpha = sigma3;
		else
			alpha = max(sigma1, -0.5*(gs/(fnew - f - gs)));

		/* Update the trust region bound according to the ratio
		of actual to predicted reduction. */
		if (actred < eta0*prered)
			delta = min(max(alpha, sigma1)*snorm, sigma2*delta);
		else if (actred < eta1*prered)
			delta = max(sigma1*delta, min(alpha*snorm, sigma2*delta));
		else if (actred < eta2*prered)
			delta = max(sigma1*delta, min(alpha*snorm, sigma3*delta));
		else
			delta = max(delta, min(alpha*snorm, sigma3*delta));
			      
		info("iter %2d act %5.3e pre %5.3e delta %5.3e f %5.3e |g| %5.3e CG %3d\n", iter, actred, prered, delta, f, g_norm, cg_iter); 

		if (actred > eta0*prered) 
		{
			iter++;
			memcpy(w, w_new, sizeof(double)*n);
			f = fnew;
		        grad(l, n, w, x_val, x_ir, x_jc, y, C, z, g, D);
			
		}
		if (f < 1.0e-32)
		{
			info("warning: f < 1.0e-32\n");
			break;
		}
		if (fabs(actred) <= 0 && fabs(prered) <= 0)
		{
			info("warning: actred and prered <= 0\n");
			break;
		}
		if (fabs(actred) <= 1.0e-12*fabs(f) && 
		    fabs(prered) <= 1.0e-12*fabs(f))
		{
			info("warning: actred and prered too small\n");
			break;
		}

	}	

	free(I);
	free(g);
	free(s);
	free(r);
	free(z);
	free(D);
	free(w_new);
	free(C);
}


void lr_train_one(const lr_problem *prob, const lr_parameter *param, double *w, double Cp, double Cn)
{
	switch(param->lr_type)
	{
		case C_SVC:
			tron(prob, param, w, Cp, Cn);
			break;
	}
}

//
// Interface functions
//
void lr_train(const lr_problem *prob, const lr_parameter *param, double *w)
{
	int i;
	double Cp, Cn;

	Cp = param->C; Cn = param->C;
	for(i=0;i<param->nr_weight;i++)
	{
		if (param->weight_label[i] == 1)
			Cp *= param->weight[i];
		else
			Cn *= param->weight[i];
	}
	lr_train_one(prob, param, w, Cp, Cn);
}

const char *lr_type_table[] =
{
	"c_svc","nu_svc","one_class","epsilon_svr","nu_svr",NULL
};

void lr_destroy_param(lr_parameter* param)
{
	free(param->weight_label);
	free(param->weight);
}

const char *lr_check_parameter(const lr_problem *prob, const lr_parameter *param)
{
	// lr_type

	int lr_type = param->lr_type;
	if(lr_type != C_SVC &&
	   lr_type != NU_SVC &&
	   lr_type != ONE_CLASS &&
	   lr_type != EPSILON_SVR &&
	   lr_type != NU_SVR)
		return "unknown lr type";

	// eps,C

	if(param->eps <= 0)
		return "eps <= 0";

	if(param->C <= 0)
		return "C <= 0";

	return NULL;
}

