#include <stdio.h>
#include <ctype.h>

#include "mex.h"
#include "matrix.h"

double bias;

// read in a problem (in svmlight format)

void read_problem(const char *filename, mxArray *plhs[])
{
	int elements, max_index, min_index, i, k;
	FILE *fp = fopen(filename,"r");
	int l = 0;
	mwIndex *ir, *jc;
	// int *ir, *jc;
	double *labels, *samples;
	
	if(fp == NULL)
	{
		fprintf(stderr,"can't open input file %s\n",filename);
		exit(1);
	}

	max_index = 0;
	min_index = 1; // our index starts from 1
	elements = 0;
	while(1)
	{
		// label
		int index;
		double value;
		fscanf(fp,"%lf",&value);

		// features
		while(1)
		{
			int c;
			do {
				c = getc(fp);
				if(c=='\n') goto out;
				if(c==EOF) goto eof;
			} while(isspace(c));
			ungetc(c,fp);
			fscanf(fp,"%d:%lf",&index, &value);
			if (index < min_index)
				min_index = index;
			elements++;
		}	
out:
		if(index > max_index)
			max_index = index;
		l++;
	}
eof:
	mexPrintf("DEBUG: elements %d l %d max_index %d min_index %d\n", elements,l,max_index, min_index);

	rewind(fp);

	// y
	plhs[0] = mxCreateDoubleMatrix(l, 1, mxREAL);
	// x^T
	if (min_index <= 0)
		plhs[1] = mxCreateSparse(max_index-min_index+1+1, l, elements+l, mxREAL);
	else
		plhs[1] = mxCreateSparse(max_index+1, l, elements+l, mxREAL);

	labels = mxGetPr(plhs[0]);
	samples = mxGetPr(plhs[1]);
	ir = mxGetIr(plhs[1]);
	jc = mxGetJc(plhs[1]);

	k=0;
	for(i=0;i<l;i++)
	{
		jc[i] = k;
		fscanf(fp,"%lf",&labels[i]);

		while(1)
		{
			int c;
			do {
				c = getc(fp);
				if(c=='\n') goto out2;
			} while(isspace(c));
			ungetc(c,fp);
			fscanf(fp,"%d:%lf",&ir[k],&samples[k]);
			if (min_index > 0)
				ir[k] -= min_index;
			++k;
		}	
out2:
		ir[k] = max_index + 1 - min_index;
		samples[k] = bias; // 1.0 for bias term feature
		++k;
	}
	jc[l] = k;

	fclose(fp);
}

void mexFunction( int nlhs, mxArray *plhs[],
		int nrhs, const mxArray *prhs[] )
{
	if(nrhs == 2)
	{
		char filename[256];

		mxGetString(prhs[0], filename, mxGetN(prhs[0]) + 1);

		if(filename == NULL)
		{
			mexPrintf("Error: filename is NULL\n");
			return;
		}

		bias = mxGetScalar(prhs[1]);
		read_problem(filename, plhs);
	}
}
